# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import os

from django import forms
from django.utils.translation import gettext_lazy as _

from feincms import settings

from . import logger
from .fields import AdminFileWithPreviewWidget
from .models import Category, MediaFile


# ------------------------------------------------------------------------
class MediaCategoryAdminForm(forms.ModelForm):
    class Meta:
        model = Category
        fields = "__all__"

    def clean_parent(self):
        data = self.cleaned_data["parent"]
        if data is not None and self.instance in data.path_list():
            raise forms.ValidationError(_("This would create a loop in the hierarchy"))

        return data

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["parent"].queryset = self.fields["parent"].queryset.exclude(
            pk=self.instance.pk
        )


# ------------------------------------------------------------------------
class MediaFileAdminForm(forms.ModelForm):
    class Meta:
        model = MediaFile
        widgets = {"file": AdminFileWithPreviewWidget}
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if settings.FEINCMS_MEDIAFILE_OVERWRITE and self.instance.id:
            field = self.instance.file.field
            if not hasattr(field, "_feincms_generate_filename_patched"):
                original_generate = field.generate_filename

                def _gen_fname(instance, filename):
                    if instance.id and hasattr(instance, "original_name"):
                        logger.info(
                            "Overwriting file %s with new data"
                            % (instance.original_name)
                        )
                        instance.file.storage.delete(instance.original_name)
                        return instance.original_name

                    return original_generate(instance, filename)

                field.generate_filename = _gen_fname
                field._feincms_generate_filename_patched = True

    def clean_file(self):
        if settings.FEINCMS_MEDIAFILE_OVERWRITE and self.instance.id:
            new_base, new_ext = os.path.splitext(self.cleaned_data["file"].name)
            old_base, old_ext = os.path.splitext(self.instance.file.name)

            if new_ext.lower() != old_ext.lower():
                raise forms.ValidationError(
                    _(
                        "Cannot overwrite with different file type (attempt to"
                        " overwrite a %(old_ext)s with a %(new_ext)s)"
                    )
                    % {"old_ext": old_ext, "new_ext": new_ext}
                )

            self.instance.original_name = self.instance.file.name

        return self.cleaned_data["file"]


# ------------------------------------------------------------------------



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# FeinCMS django-tagging support. To add tagging to your (page) model,
# simply do a
#
#    from feincms.contrib import tagging
#    tagging.tag_model(Page)
# ------------------------------------------------------------------------


from django import forms
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.db.models.signals import pre_save
from django.utils.translation import gettext_lazy as _
from tagging.fields import TagField
from tagging.models import Tag
from tagging.registry import AlreadyRegistered, register as tagging_register
from tagging.utils import parse_tag_input


# ------------------------------------------------------------------------
def taglist_to_string(taglist):
    retval = ""
    if len(taglist) >= 1:
        taglist.sort()
        retval = ",".join(taglist)
    return retval


# ------------------------------------------------------------------------
# The following is lifted from:
# http://code.google.com/p/django-tagging/issues/detail?id=189


"""
TagSelectField

A variation of the django-tagging TagField which uses a
SelectMultiple widget instead of free text field.

class MyModel(models.Model):
    ...
    tags = TagSelectField(filter_horizontal=True, blank=False)

"""


class TagSelectFormField(forms.MultipleChoiceField):
    def clean(self, value):
        return taglist_to_string(list(value))


class Tag_formatvalue_mixin:
    def format_value(self, value):
        value = parse_tag_input(value or "")
        return super().format_value(value)


class fv_FilteredSelectMultiple(Tag_formatvalue_mixin, FilteredSelectMultiple):
    pass


class fv_SelectMultiple(Tag_formatvalue_mixin, forms.SelectMultiple):
    pass


class TagSelectField(TagField):
    def __init__(self, filter_horizontal=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_horizontal = filter_horizontal

    def formfield(self, **defaults):
        if self.filter_horizontal:
            widget = fv_FilteredSelectMultiple(self.verbose_name, is_stacked=False)
        else:
            widget = fv_SelectMultiple()

        defaults["widget"] = widget
        choices = [(str(t), str(t)) for t in Tag.objects.all()]
        return TagSelectFormField(choices=choices, required=not self.blank, **defaults)


# ------------------------------------------------------------------------
def pre_save_handler(sender, instance, **kwargs):
    """
    Intercept attempts to save and sort the tag field alphabetically, so
    we won't have different permutations in the filter list.
    """
    taglist = parse_tag_input(instance.tags)
    instance.tags = taglist_to_string(taglist)


# ------------------------------------------------------------------------
def tag_model(
    cls,
    admin_cls=None,
    field_name="tags",
    sort_tags=False,
    select_field=False,
    auto_add_admin_field=True,
    admin_list_display=True,
):
    """
    tag_model accepts a number of named parameters:

    admin_cls   If set to a subclass of ModelAdmin, will insert the tag
                field into the list_display and list_filter fields.
    field_name  Defaults to "tags", can be used to name your tag field
                differently.
    sort_tags   Boolean, defaults to False. If set to True, a pre_save
                handler will be inserted to sort the tag field alphabetically.
                This is useful in case you want a canonical representation
                for a tag collection, as when yo're presenting a list of
                tag combinations (e.g. in an admin filter list).
    select_field If True, show a multi select instead of the standard
                CharField for tag entry.
    auto_add_admin_field If True, attempts to add the tag field to the admin
                class.
    """

    cls.add_to_class(
        field_name,
        (TagSelectField if select_field else TagField)(
            field_name.capitalize(), blank=True
        ),
    )
    # use another name for the tag descriptor
    # See http://code.google.com/p/django-tagging/issues/detail?id=95 for the
    # reason why
    try:
        tagging_register(cls, tag_descriptor_attr="tagging_" + field_name)
    except AlreadyRegistered:
        return

    if admin_cls:
        if admin_list_display:
            admin_cls.list_display.append(field_name)
        admin_cls.list_filter.append(field_name)

        if auto_add_admin_field and hasattr(admin_cls, "add_extension_options"):
            admin_cls.add_extension_options(_("Tagging"), {"fields": (field_name,)})

    if sort_tags:
        pre_save.connect(pre_save_handler, sender=cls)


# ------------------------------------------------------------------------



def __getattr__(key):
    # Work around Django 3.2's autoloading of *.apps modules (AppConfig
    # autodiscovery)
    if key in {
        "ApplicationContent",
        "app_reverse",
        "app_reverse_lazy",
        "permalink",
        "UnpackTemplateResponse",
        "standalone",
        "unpack",
    }:
        from feincms.content.application import models

        return getattr(models, key)

    raise AttributeError("Unknown attribute '%s'" % key)



from django.http import Http404
from django.shortcuts import get_object_or_404

from feincms.views import Handler


class PreviewHandler(Handler):
    """
    Preview handler

    The methods used in this handler should not be considered official API.

    *** Everything here is subject to change. ***
    """

    def get_object(self):
        """Get the page by the id in the url here instead."""

        # This might happen when there is a preview request for
        # a page that doesn't exist, and now FeinCMS wants to
        # show a 404 page via FEINCMS_CMS_404_PAGE.
        if len(self.args) < 2:
            return super().get_object()

        page = get_object_or_404(self.page_model, pk=self.args[1])
        # Remove _preview/42/ from URL, the rest of the handler code should not
        # know that anything about previewing. Handler.prepare will still raise
        # a 404 if the extra_path isn't consumed by any content type
        self.request.path = page.get_absolute_url()

        return page

    def handler(self, request, *args, **kwargs):
        if not request.user.is_staff:
            raise Http404("Not found (not allowed)")
        response = super().handler(request, *args, **kwargs)
        response["Cache-Control"] = "no-cache, must-revalidate, no-store, private"
        return response



"""
Add a keyword and a description field which are helpful for SEO optimization.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "meta_keywords",
            models.TextField(
                _("meta keywords"),
                blank=True,
                help_text=_("Keywords are ignored by most search engines."),
            ),
        )
        self.model.add_to_class(
            "meta_description",
            models.TextField(
                _("meta description"),
                blank=True,
                help_text=_(
                    "This text is displayed on the search results page. "
                    "It is however not used for the SEO ranking. "
                    "Text longer than 140 characters is truncated."
                ),
            ),
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.extend_list("search_fields", ["meta_keywords", "meta_description"])

        modeladmin.add_extension_options(
            _("Search engine optimization"),
            {"fields": ("meta_keywords", "meta_description"), "classes": ("collapse",)},
        )



from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms.admin.item_editor import FeinCMSInline
from feincms.module.medialibrary.fields import ContentWithMediaFile
from feincms.utils.tuple import AutoRenderTuple


class MediaFileContentInline(FeinCMSInline):
    raw_id_fields = ("mediafile",)
    radio_fields = {"type": admin.VERTICAL}


class MediaFileContent(ContentWithMediaFile):
    """
    Rehashed, backwards-incompatible media file content which does not contain
    the problems from v1 anymore.

    Create a media file content as follows::

        from feincms.content.medialibrary.models import MediaFileContent
        Page.create_content_type(MediaFileContent, TYPE_CHOICES=(
            ('default', _('Default')),
            ('lightbox', _('Lightbox')),
            ('whatever', _('Whatever')),
            ))

    For a media file of type 'image' and type 'lightbox', the following
    templates are tried in order:

    * content/mediafile/image_lightbox.html
    * content/mediafile/image.html
    * content/mediafile/lightbox.html
    * content/mediafile/default.html

    The context contains ``content`` and ``request`` (if available).
    """

    feincms_item_editor_inline = MediaFileContentInline

    class Meta:
        abstract = True
        verbose_name = _("media file")
        verbose_name_plural = _("media files")

    @classmethod
    def initialize_type(cls, TYPE_CHOICES=None):
        if TYPE_CHOICES is None:
            raise ImproperlyConfigured(
                "You have to set TYPE_CHOICES when" " creating a %s" % cls.__name__
            )

        cls.add_to_class(
            "type",
            models.CharField(
                _("type"),
                max_length=20,
                choices=TYPE_CHOICES,
                default=TYPE_CHOICES[0][0],
            ),
        )

    def render(self, **kwargs):
        return AutoRenderTuple(
            (
                [
                    f"content/mediafile/{self.mediafile.type}_{self.type}.html",
                    "content/mediafile/%s.html" % self.mediafile.type,
                    "content/mediafile/%s.html" % self.type,
                    "content/mediafile/default.html",
                ],
                {"content": self},
            )
        )



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


from django.contrib import admin
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured

from feincms import settings

from .modeladmins import PageAdmin
from .models import Page


# ------------------------------------------------------------------------

if settings.FEINCMS_USE_PAGE_ADMIN:
    try:
        Page._meta.get_field("template_key")
    except FieldDoesNotExist:
        raise ImproperlyConfigured(
            "The page module requires a 'Page.register_templates()' call "
            "somewhere ('Page.register_regions()' is not sufficient). "
            "If you're not using the default Page admin, maybe try "
            "FEINCMS_USE_PAGE_ADMIN=False to avoid this warning."
        )

    admin.site.register(Page, PageAdmin)

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import logging

from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

from feincms.utils import get_singleton, get_singleton_url


register = template.Library()


def _render_content(content, **kwargs):
    # Track current render level and abort if we nest too deep. Avoids
    # crashing in recursive page contents (eg. a page list that contains
    # itself or similar).
    request = kwargs.get("request")
    if request is not None:
        level = getattr(request, "feincms_render_level", 0)
        if level > 10:
            logging.getLogger("feincms").error(
                f"Refusing to render {content!r}, render level is already {level}"
            )
            return
        setattr(request, "feincms_render_level", level + 1)

    r = content.render(**kwargs)

    if request is not None:
        level = getattr(request, "feincms_render_level", 1)
        setattr(request, "feincms_render_level", max(level - 1, 0))

    if isinstance(r, (list, tuple)):
        # Modeled after feincms3's TemplatePluginRenderer
        context = kwargs["context"]
        plugin_template, plugin_context = r

        if not hasattr(plugin_template, "render"):  # Quacks like a template?
            try:
                engine = context.template.engine
            except AttributeError:
                from django.template.engine import Engine

                engine = Engine.get_default()

            if isinstance(plugin_template, (list, tuple)):
                plugin_template = engine.select_template(plugin_template)
            else:
                plugin_template = engine.get_template(plugin_template)

        with context.push(plugin_context):
            return plugin_template.render(context)

    return r


@register.simple_tag(takes_context=True)
def feincms_render_region(context, feincms_object, region, request=None):
    """
    {% feincms_render_region feincms_page "main" request %}
    """
    if not feincms_object:
        return ""

    return mark_safe(
        "".join(
            _render_content(content, request=request, context=context)
            for content in getattr(feincms_object.content, region)
        )
    )


@register.simple_tag(takes_context=True)
def feincms_render_content(context, content, request=None):
    """
    {% feincms_render_content content request %}
    """
    if not content:
        return ""

    return _render_content(content, request=request, context=context)


@register.simple_tag
def feincms_load_singleton(template_key, cls=None):
    """
    {% feincms_load_singleton template_key %} -- return a FeinCMS
    Base object which uses a Template with singleton=True.
    """
    return get_singleton(template_key, cls, raise_exception=settings.DEBUG)


@register.simple_tag
def feincms_singleton_url(template_key, cls=None):
    """
    {% feincms_singleton_url template_key %} -- return the URL of a FeinCMS
    Base object which uses a Template with singleton=True.
    """
    return get_singleton_url(template_key, cls, raise_exception=settings.DEBUG)



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


from django.contrib.admin.widgets import AdminFileWidget, ForeignKeyRawIdWidget
from django.db import models
from django.utils.html import escape, mark_safe
from django.utils.translation import gettext_lazy as _

from feincms.admin.item_editor import FeinCMSInline
from feincms.utils import shorten_string

from .models import MediaFile
from .thumbnail import admin_thumbnail


__all__ = ("MediaFileForeignKey", "ContentWithMediaFile")


# ------------------------------------------------------------------------
class MediaFileForeignKeyRawIdWidget(ForeignKeyRawIdWidget):
    def __init__(self, original):
        self.__dict__ = original.__dict__

    def label_and_url_for_value(self, value):
        label, url = super().label_and_url_for_value(value)
        key = "pk"
        try:
            obj = (
                self.rel.model._default_manager.using(self.db)
                .filter(**{key: value})
                .first()
            )
            label = ["&nbsp;<strong>%s</strong>" % escape(shorten_string(str(obj)))]
            image = admin_thumbnail(obj)

            if image:
                label.append(
                    '<br /><img src="%s" alt="" style="margin:1em 0 0 170px"'
                    "/>" % image
                )

            return mark_safe("".join(label)), url
        except (ValueError, self.rel.model.DoesNotExist):
            return label, url


class MediaFileForeignKey(models.ForeignKey):
    """
    Drop-in replacement for Django's ``models.ForeignKey`` which automatically
    adds a thumbnail of media files if the media file foreign key is shown
    using ``raw_id_fields``.
    """

    def __init__(self, *args, **kwargs):
        if not args and "to" not in kwargs:
            args = (MediaFile,)
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        if "widget" in kwargs and isinstance(kwargs["widget"], ForeignKeyRawIdWidget):
            kwargs["widget"] = MediaFileForeignKeyRawIdWidget(kwargs["widget"])
        return super().formfield(**kwargs)


class ContentWithMediaFile(models.Model):
    class feincms_item_editor_inline(FeinCMSInline):
        raw_id_fields = ("mediafile",)

    mediafile = MediaFileForeignKey(
        MediaFile,
        verbose_name=_("media file"),
        related_name="+",
        on_delete=models.PROTECT,
    )

    class Meta:
        abstract = True


# ------------------------------------------------------------------------
class AdminFileWithPreviewWidget(AdminFileWidget):
    """
    Simple AdminFileWidget, but detects if the file is an image and
    tries to render a small thumbnail besides the input field.
    """

    def render(self, name, value, attrs=None, *args, **kwargs):
        r = super().render(name, value, attrs=attrs, *args, **kwargs)

        if value and getattr(value, "instance", None):
            image = admin_thumbnail(value.instance)
            if image:
                r = mark_safe(
                    (
                        '<img src="%s" alt="" style="float: left; padding-right:'
                        '8px; border-right: 1px solid #ccc; margin-right: 8px"'
                        ">" % image
                    )
                    + r
                )

        return r



from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms.content.raw.models import RawContent  # noqa
from feincms.content.richtext.models import RichTextContent  # noqa
from feincms.utils.tuple import AutoRenderTuple


class TemplateContent(models.Model):
    """
    Pass a list of templates when creating this content type. It uses the
    default template system::

        Page.create_content_type(TemplateContent, TEMPLATES=[
            ('content/template/something1.html', 'something'),
            ('content/template/something2.html', 'something else'),
            ('base.html', 'makes no sense'),
        ])
    """

    class Meta:
        abstract = True
        verbose_name = _("template content")
        verbose_name_plural = _("template contents")

    @classmethod
    def initialize_type(cls, TEMPLATES):
        cls.add_to_class(
            "template",
            models.CharField(_("template"), max_length=100, choices=TEMPLATES),
        )

    def render(self, **kwargs):
        return AutoRenderTuple((self.template, {"content": self}))



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import os

from django import forms
from django.conf import settings as django_settings
from django.contrib import admin, messages
from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
from django.contrib.auth.decorators import permission_required
from django.contrib.sites.shortcuts import get_current_site
from django.core.files.images import get_image_dimensions
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.template.defaultfilters import filesizeformat
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, ngettext
from django.views.decorators.csrf import csrf_protect

from feincms.extensions import ExtensionModelAdmin
from feincms.translations import admin_translationinline, lookup_translations
from feincms.utils import shorten_string

from .forms import MediaCategoryAdminForm, MediaFileAdminForm
from .models import Category, MediaFileTranslation
from .thumbnail import admin_thumbnail
from .zip import import_zipfile


# -----------------------------------------------------------------------
class CategoryAdmin(admin.ModelAdmin):
    form = MediaCategoryAdminForm
    list_display = ["path"]
    list_filter = ["parent"]
    list_per_page = 25
    search_fields = ["title"]
    prepopulated_fields = {"slug": ("title",)}


# ------------------------------------------------------------------------
@admin.action(description=_("Add selected media files to category"))
def assign_category(modeladmin, request, queryset):
    class AddCategoryForm(forms.Form):
        _selected_action = forms.CharField(widget=forms.MultipleHiddenInput)
        category = forms.ModelChoiceField(Category.objects.all())

    form = None
    if "apply" in request.POST:
        form = AddCategoryForm(request.POST)
        if form.is_valid():
            category = form.cleaned_data["category"]

            count = 0
            for mediafile in queryset:
                category.mediafile_set.add(mediafile)
                count += 1

            message = ngettext(
                "Successfully added %(count)d media file to %(category)s.",
                "Successfully added %(count)d media files to %(category)s.",
                count,
            ) % {"count": count, "category": category}
            modeladmin.message_user(request, message)
            return HttpResponseRedirect(request.get_full_path())
    if "cancel" in request.POST:
        return HttpResponseRedirect(request.get_full_path())

    if not form:
        form = AddCategoryForm(
            initial={"_selected_action": request.POST.getlist(ACTION_CHECKBOX_NAME)}
        )

    return render(
        request,
        "admin/medialibrary/add_to_category.html",
        {"mediafiles": queryset, "category_form": form, "opts": modeladmin.model._meta},
    )


# -------------------------------------------------------------------------
@admin.action(description=_("Export selected media files as zip file"))
def save_as_zipfile(modeladmin, request, queryset):
    from .zip import export_zipfile

    site = get_current_site(request)
    try:
        zip_name = export_zipfile(site, queryset)
        messages.info(request, _("ZIP file exported as %s") % zip_name)
    except Exception as e:
        messages.error(request, _("ZIP file export failed: %s") % str(e))
        return

    return HttpResponseRedirect(os.path.join(django_settings.MEDIA_URL, zip_name))


# ------------------------------------------------------------------------
class MediaFileAdmin(ExtensionModelAdmin):
    form = MediaFileAdminForm

    save_on_top = True
    date_hierarchy = "created"
    inlines = [admin_translationinline(MediaFileTranslation)]
    list_display = ["admin_thumbnail", "__str__", "file_info", "formatted_created"]
    list_display_links = ["__str__"]
    list_filter = ["type", "categories"]
    list_per_page = 25
    search_fields = ["copyright", "file", "translations__caption"]
    filter_horizontal = ("categories",)
    actions = [assign_category, save_as_zipfile]

    def get_urls(self):
        from django.urls import path

        return [
            path(
                "mediafile-bulk-upload/",
                self.admin_site.admin_view(MediaFileAdmin.bulk_upload),
                {},
                name="mediafile_bulk_upload",
            )
        ] + super().get_urls()

    def changelist_view(self, request, extra_context=None):
        if extra_context is None:
            extra_context = {}
        extra_context["categories"] = Category.objects.order_by("title")
        return super().changelist_view(request, extra_context=extra_context)

    @admin.display(description=_("Preview"))
    def admin_thumbnail(self, obj):
        image = admin_thumbnail(obj)
        if image:
            return mark_safe(
                """
                <a href="%(url)s" target="_blank">
                    <img src="%(image)s" alt="" />
                </a>"""
                % {"url": obj.file.url, "image": image}
            )
        return ""

    @admin.display(
        description=_("file size"),
        ordering="file_size",
    )
    def formatted_file_size(self, obj):
        return filesizeformat(obj.file_size)

    @admin.display(
        description=_("created"),
        ordering="created",
    )
    def formatted_created(self, obj):
        return obj.created.strftime("%Y-%m-%d")

    @admin.display(
        description=_("file type"),
        ordering="type",
    )
    def file_type(self, obj):
        t = obj.filetypes_dict[obj.type]
        if obj.type == "image":
            # get_image_dimensions is expensive / slow if the storage is not
            # local filesystem (indicated by availability the path property)
            try:
                obj.file.path
            except NotImplementedError:
                return t
            try:
                d = get_image_dimensions(obj.file.file, close=True)
                if d:
                    t += " %d&times;%d" % (d[0], d[1])
            except (OSError, TypeError, ValueError) as e:
                t += " (%s)" % e
        return mark_safe(t)

    @admin.display(
        description=_("file info"),
        ordering="file",
    )
    def file_info(self, obj):
        """
        Method for showing the file name in admin.

        Note: This also includes a hidden field that can be used to extract
        the file name later on, this can be used to access the file name from
        JS, like for example a TinyMCE connector shim.
        """
        return mark_safe(
            (
                '<input type="hidden" class="medialibrary_file_path"'
                ' name="_media_path_%d" value="%s" id="_refkey_%d" />'
                " %s <br />%s, %s"
            )
            % (
                obj.id,
                obj.file.name,
                obj.id,
                shorten_string(os.path.basename(obj.file.name), max_length=40),
                self.file_type(obj),
                self.formatted_file_size(obj),
            )
        )

    @staticmethod
    @csrf_protect
    @permission_required("medialibrary.add_mediafile")
    def bulk_upload(request):
        if request.method == "POST" and "data" in request.FILES:
            try:
                count = import_zipfile(
                    request.POST.get("category"),
                    request.POST.get("overwrite", False),
                    request.FILES["data"],
                )
                messages.info(request, _("%d files imported") % count)
            except Exception as e:
                messages.error(request, _("ZIP import failed: %s") % e)
        else:
            messages.error(request, _("No input file given"))

        return HttpResponseRedirect(reverse("admin:medialibrary_mediafile_changelist"))

    def get_queryset(self, request):
        return super().get_queryset(request).transform(lookup_translations())

    def save_model(self, request, obj, form, change):
        if obj.id:
            obj.purge_translation_cache()
        return super().save_model(request, obj, form, change)


# ------------------------------------------------------------------------



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import re

from django.apps import apps
from django.contrib.admin.widgets import ForeignKeyRawIdWidget
from django.forms.models import model_to_dict
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from mptt.forms import MPTTAdminForm


class RedirectToWidget(ForeignKeyRawIdWidget):
    def label_for_value(self, value):
        match = re.match(
            # XXX this regex would be available as .models.REDIRECT_TO_RE
            r"^(?P<app_label>\w+).(?P<model_name>\w+):(?P<pk>\d+)$",
            value,
        )

        if match:
            matches = match.groupdict()
            model = apps.get_model(matches["app_label"], matches["model_name"])
            try:
                instance = model._default_manager.get(pk=int(matches["pk"]))
                return "&nbsp;<strong>{} ({})</strong>".format(
                    instance,
                    instance.get_absolute_url(),
                )

            except model.DoesNotExist:
                pass

        return ""


# ------------------------------------------------------------------------
class PageAdminForm(MPTTAdminForm):
    never_copy_fields = (
        "title",
        "slug",
        "parent",
        "active",
        "override_url",
        "translation_of",
        "_content_title",
        "_page_title",
    )

    @property
    def page_model(self):
        return self._meta.model

    @property
    def page_manager(self):
        return self.page_model._default_manager

    def __init__(self, *args, **kwargs):
        if "initial" in kwargs:
            if "parent" in kwargs["initial"]:
                # Prefill a few form values from the parent page
                try:
                    page = self.page_manager.get(pk=kwargs["initial"]["parent"])

                    data = model_to_dict(page)

                    for field in self.page_manager.exclude_from_copy:
                        if field in data:
                            del data[field]

                    # These are always excluded from prefilling
                    for field in self.never_copy_fields:
                        if field in data:
                            del data[field]

                    data.update(kwargs["initial"])
                    if page.template.child_template:
                        data["template_key"] = page.template.child_template
                    kwargs["initial"] = data
                except self.page_model.DoesNotExist:
                    pass

            elif "translation_of" in kwargs["initial"]:
                # Only if translation extension is active
                try:
                    page = self.page_manager.get(pk=kwargs["initial"]["translation_of"])
                    original = page.original_translation

                    data = {
                        "translation_of": original.id,
                        "template_key": original.template_key,
                        "active": original.active,
                        "in_navigation": original.in_navigation,
                    }

                    if original.parent:
                        try:
                            data["parent"] = original.parent.get_translation(
                                kwargs["initial"]["language"]
                            ).id
                        except self.page_model.DoesNotExist:
                            # ignore this -- the translation does not exist
                            pass

                    data.update(kwargs["initial"])
                    kwargs["initial"] = data
                except (AttributeError, self.page_model.DoesNotExist):
                    pass

        # Not required, only a nice-to-have for the `redirect_to` field
        modeladmin = kwargs.pop("modeladmin", None)
        super().__init__(*args, **kwargs)
        if modeladmin and "redirect_to" in self.fields:
            # Note: Using `parent` is not strictly correct, but we can be
            # sure that `parent` always points to another page instance,
            # and that's good enough for us.
            field = self.page_model._meta.get_field("parent")
            self.fields["redirect_to"].widget = RedirectToWidget(
                field.remote_field
                if hasattr(field, "remote_field")
                else field.rel,  # noqa
                modeladmin.admin_site,
            )

        if "template_key" in self.fields:
            choices = []
            for key, template_name in self.page_model.TEMPLATE_CHOICES:
                template = self.page_model._feincms_templates[key]
                pages_for_template = self.page_model._default_manager.filter(
                    template_key=key
                )
                pk = kwargs["instance"].pk if kwargs.get("instance") else None
                other_pages_for_template = pages_for_template.exclude(pk=pk)
                if template.singleton and other_pages_for_template.exists():
                    continue  # don't allow selection of singleton if in use
                if template.preview_image:
                    choices.append(
                        (
                            template.key,
                            mark_safe(
                                '<img src="%s" alt="%s" /> %s'
                                % (template.preview_image, template.key, template.title)
                            ),
                        )
                    )
                else:
                    choices.append((template.key, template.title))

            self.fields["template_key"].choices = choices

    def clean(self):
        cleaned_data = super().clean()

        # No need to think further, let the user correct errors first
        if self._errors:
            return cleaned_data

        current_id = None
        # See the comment below on why we do not use Page.objects.active(),
        # at least for now.
        active_pages = self.page_manager.filter(active=True)

        if self.instance:
            current_id = self.instance.id
            active_pages = active_pages.exclude(id=current_id)

        sites_is_installed = apps.is_installed("django.contrib.sites")
        if sites_is_installed and "site" in cleaned_data:
            active_pages = active_pages.filter(site=cleaned_data["site"])

        # Convert PK in redirect_to field to something nicer for the future
        redirect_to = cleaned_data.get("redirect_to")
        if redirect_to and re.match(r"^\d+$", redirect_to):
            opts = self.page_model._meta
            cleaned_data["redirect_to"] = "{}.{}:{}".format(
                opts.app_label,
                opts.model_name,
                redirect_to,
            )

        if "active" in cleaned_data and not cleaned_data["active"]:
            # If the current item is inactive, we do not need to conduct
            # further validation. Note that we only check for the flag, not
            # for any other active filters. This is because we do not want
            # to inspect the active filters to determine whether two pages
            # really won't be active at the same time.
            return cleaned_data

        if "override_url" in cleaned_data and cleaned_data["override_url"]:
            if active_pages.filter(_cached_url=cleaned_data["override_url"]).count():
                self._errors["override_url"] = self.error_class(
                    [_("This URL is already taken by an active page.")]
                )
                del cleaned_data["override_url"]

            return cleaned_data

        if current_id:
            # We are editing an existing page
            parent = self.page_manager.get(pk=current_id).parent
        else:
            # The user tries to create a new page
            parent = cleaned_data["parent"]

        if parent:
            new_url = "{}{}/".format(parent._cached_url, cleaned_data["slug"])
        else:
            new_url = "/%s/" % cleaned_data["slug"]

        if active_pages.filter(_cached_url=new_url).count():
            self._errors["active"] = self.error_class(
                [_("This URL is already taken by another active page.")]
            )
            del cleaned_data["active"]

        if parent and parent.template.enforce_leaf:
            self._errors["parent"] = self.error_class(
                [_("This page does not allow attachment of child pages")]
            )
            del cleaned_data["parent"]

        return cleaned_data


# ------------------------------------------------------------------------
# ------------------------------------------------------------------------



from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.contrib.richtext import RichTextField
from feincms.utils.tuple import AutoRenderTuple


class RichTextContent(models.Model):
    """
    Rich text content. Uses TinyMCE by default, but can be configured to do
    anything you want using ``FEINCMS_RICHTEXT_INIT_CONTEXT`` and
    ``FEINCMS_RICHTEXT_INIT_TEMPLATE``.

    If you are using TinyMCE 4.x then ``FEINCMS_RICHTEXT_INIT_TEMPLATE``
    needs to be set to ``admin/content/richtext/init_tinymce4.html``.

    Optionally runs the HTML code through HTML cleaners if you specify
    ``cleanse=True`` when calling ``create_content_type``.
    """

    feincms_item_editor_context_processors = (
        lambda x: settings.FEINCMS_RICHTEXT_INIT_CONTEXT,
    )
    feincms_item_editor_includes = {"head": [settings.FEINCMS_RICHTEXT_INIT_TEMPLATE]}

    class Meta:
        abstract = True
        verbose_name = _("rich text")
        verbose_name_plural = _("rich texts")

    def render(self, **kwargs):
        return AutoRenderTuple(("content/richtext/default.html", {"content": self}))

    @classmethod
    def initialize_type(cls, cleanse=None):
        cls.add_to_class("text", RichTextField(_("text"), blank=True, cleanse=cleanse))



import logging
import re
import sys

from django.conf import settings as django_settings
from django.http import Http404, HttpResponseRedirect
from django.views.decorators.http import condition


logger = logging.getLogger(__name__)


def redirect_request_processor(page, request):
    """
    Returns a ``HttpResponseRedirect`` instance if the current page says
    a redirect should happen.
    """
    target = page.get_redirect_to_target(request)
    if target:
        extra_path = request._feincms_extra_context.get("extra_path", "/")
        if extra_path == "/":
            return HttpResponseRedirect(target)
        logger.debug(
            "Page redirect on '%s' not taken because extra path '%s' present",
            page.get_absolute_url(),
            extra_path,
        )
        raise Http404()


def extra_context_request_processor(page, request):
    """
    Fills ``request._feincms_extra_context`` with a few useful variables.
    """
    request._feincms_extra_context.update(
        {
            # XXX This variable name isn't accurate anymore.
            "in_appcontent_subpage": False,
            "extra_path": "/",
        }
    )

    url = page.get_absolute_url()
    if request.path != url:
        request._feincms_extra_context.update(
            {
                "in_appcontent_subpage": True,
                "extra_path": re.sub(
                    "^" + re.escape(url.rstrip("/")), "", request.path
                ),
            }
        )


class __DummyResponse(dict):
    """
    This is a dummy class with enough behaviour of HttpResponse so we
    can use the condition decorator without too much pain.
    """

    @property
    def headers(self):
        return self

    def has_header(self, what):
        return False


def etag_request_processor(page, request):
    """
    Short-circuits the request-response cycle if the ETag matches.
    """

    def dummy_response_handler(*args, **kwargs):
        return __DummyResponse()

    def etagger(request, page, *args, **kwargs):
        etag = page.etag(request)
        return etag

    def lastmodifier(request, page, *args, **kwargs):
        lm = page.last_modified()
        return lm

    # Now wrap the condition decorator around our dummy handler:
    # the net effect is that we will be getting a DummyResponse from
    # the handler if processing is to continue and a non-DummyResponse
    # (should be a "304 not modified") if the etag matches.
    rsp = condition(etag_func=etagger, last_modified_func=lastmodifier)(
        dummy_response_handler
    )(request, page)

    # If dummy then don't do anything, if a real response, return and
    # thus shortcut the request processing.
    if not isinstance(rsp, __DummyResponse):
        return rsp


def etag_response_processor(page, request, response):
    """
    Response processor to set an etag header on outgoing responses.
    The Page.etag() method must return something valid as etag content
    whenever you want an etag header generated.
    """
    etag = page.etag(request)
    if etag is not None:
        response["ETag"] = '"' + etag + '"'


def debug_sql_queries_response_processor(verbose=False, file=sys.stderr):
    """
    Attaches a handler which prints the query count (and optionally all
    individual queries which have been executed) on the console. Does nothing
    if ``DEBUG = False``.

    Example::

        from feincms.module.page import models, processors
        models.Page.register_response_processor(
            processors.debug_sql_queries_response_processor(verbose=True),
            )
    """
    if not django_settings.DEBUG:
        return lambda page, request, response: None

    def processor(page, request, response):
        from django.db import connection

        try:
            import sqlparse

            def print_sql(x):
                return sqlparse.format(x, reindent=True, keyword_case="upper")

        except Exception:

            def print_sql(x):
                return x

        if verbose:
            print("-" * 60, file=file)
        time = 0.0
        i = 0
        for q in connection.queries:
            i += 1
            if verbose:
                print(
                    "%d : [%s]\n%s\n" % (i, q["time"], print_sql(q["sql"])), file=file
                )
            time += float(q["time"])

        print("-" * 60, file=file)
        print("Total: %d queries, %.3f ms" % (i, time), file=file)
        print("-" * 60, file=file)

    return processor



"""
Page navigation groups allow assigning pages to differing navigation lists
such as header, footer and what else.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    ident = "navigationgroups"
    groups = [("default", _("Default")), ("footer", _("Footer"))]

    def handle_model(self):
        self.model.add_to_class(
            "navigation_group",
            models.CharField(
                _("navigation group"),
                choices=self.groups,
                default=self.groups[0][0],
                max_length=20,
                blank=True,
                db_index=True,
            ),
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options("navigation_group")
        modeladmin.extend_list("list_display", ["navigation_group"])
        modeladmin.extend_list("list_filter", ["navigation_group"])



"""
This introduces a new page type, which has no content of its own but inherits
all content from the linked page.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions
from feincms._internal import monkeypatch_property


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "symlinked_page",
            models.ForeignKey(
                "self",
                blank=True,
                null=True,
                on_delete=models.CASCADE,
                related_name="%(app_label)s_%(class)s_symlinks",
                verbose_name=_("symlinked page"),
                help_text=_("All content is inherited from this page if given."),
            ),
        )

        @monkeypatch_property(self.model)
        def content(self):
            if not hasattr(self, "_content_proxy"):
                if self.symlinked_page:
                    self._content_proxy = self.content_proxy_class(self.symlinked_page)
                else:
                    self._content_proxy = self.content_proxy_class(self)

            return self._content_proxy

    def handle_modeladmin(self, modeladmin):
        modeladmin.extend_list("raw_id_fields", ["symlinked_page"])
        modeladmin.add_extension_options("symlinked_page")



"""
Add a many-to-many relationship field to relate this page to other pages.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions, settings


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "related_pages",
            models.ManyToManyField(
                settings.FEINCMS_DEFAULT_PAGE_MODEL,
                blank=True,
                related_name="%(app_label)s_%(class)s_related",
                help_text=_("Select pages that should be listed as related content."),
            ),
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.extend_list("filter_horizontal", ["related_pages"])

        modeladmin.add_extension_options(
            _("Related pages"), {"fields": ("related_pages",), "classes": ("collapse",)}
        )



from django.contrib.admin.filters import FieldListFilter

from .filters import CategoryFieldListFilter, ParentFieldListFilter


FieldListFilter.register(
    lambda f: getattr(f, "parent_filter", False),
    ParentFieldListFilter,
    take_priority=True,
)
FieldListFilter.register(
    lambda f: getattr(f, "category_filter", False),
    CategoryFieldListFilter,
    take_priority=True,
)



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import logging
import re
from io import BytesIO

from django import template
from django.core.cache import cache
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.utils.encoding import force_str
from PIL import Image

from feincms import settings


logger = logging.getLogger("feincms.templatetags.thumbnail")
register = template.Library()


class Thumbnailer:
    THUMBNAIL_SIZE_RE = re.compile(r"^(?P<w>\d+)x(?P<h>\d+)$")
    MARKER = "_thumb_"
    BROWSER_SUPPORTED_FORMATS = (
        "jpg",
        "jpeg",
        "png",
        "webp",
    )  # browser supported image formats
    TRANSPARENCY_SUPPORTING_FORMATS = ("png", "webp")  # formats with alpha channel

    def __init__(self, filename, size="200x200"):
        self.filename = filename
        self.size = size

    @property
    def url(self):
        return str(self)

    def __str__(self):
        match = self.THUMBNAIL_SIZE_RE.match(self.size)
        if not (self.filename and match):
            return ""

        matches = match.groupdict()

        # figure out storage
        if hasattr(self.filename, "storage"):
            storage = self.filename.storage
        else:
            storage = default_storage

        # figure out name
        if hasattr(self.filename, "name"):
            filename = self.filename.name
        else:
            filename = force_str(self.filename)

        # defining the filename and the miniature filename
        try:
            basename, format = filename.rsplit(".", 1)
        except ValueError:
            basename, format = filename, "jpg"

        miniature = "".join(
            [
                settings.FEINCMS_THUMBNAIL_DIR,
                basename,
                self.MARKER,
                self.size,
                ".",
                format,
            ]
        )

        if settings.FEINCMS_THUMBNAIL_CACHE_TIMEOUT != 0:
            cache_key = "thumb_url_%s" % miniature
            url = cache.get(cache_key)
            if url:
                return url

        if not storage.exists(miniature):
            generate = True
        else:
            try:
                generate = storage.modified_time(miniature) < storage.modified_time(
                    filename
                )
            except (NotImplementedError, AttributeError):
                # storage does NOT support modified_time
                generate = False
            except OSError:
                # Someone might have deleted the file
                return ""

        if generate:
            try:
                self.generate(
                    storage=storage,
                    original=filename,
                    size=matches,
                    miniature=miniature,
                )
            except Exception as exc:
                logger.exception("Rendering a thumbnail failed: %s", exc)
                # PIL raises a plethora of Exceptions if reading the image
                # is not possible. Since we cannot be sure what Exception will
                # happen, catch them all so the thumbnailer will never fail.
                return storage.url(filename)

        url = storage.url(miniature)
        if settings.FEINCMS_THUMBNAIL_CACHE_TIMEOUT != 0:
            cache.set(cache_key, url, timeout=settings.FEINCMS_THUMBNAIL_CACHE_TIMEOUT)
        return url

    def generate(self, storage, original, size, miniature):
        with storage.open(original) as original_handle:
            with BytesIO(original_handle.read()) as original_bytes:
                image = Image.open(original_bytes)

                # defining the size
                w, h = int(size["w"]), int(size["h"])

                assert image.format is not None
                format = image.format.lower()  # Save format for the save() call later
                image.thumbnail([w, h], Image.Resampling.LANCZOS)
                buf = BytesIO()
                if (
                    format not in self.BROWSER_SUPPORTED_FORMATS
                ):  # browser supported image formats
                    format = "jpeg"
                if (
                    format in self.TRANSPARENCY_SUPPORTING_FORMATS
                    and image.mode not in ("RGBA", "RGB", "L")
                ):
                    image = image.convert("RGBA")
                elif (
                    format not in self.TRANSPARENCY_SUPPORTING_FORMATS
                    and image.mode not in ("RGB", "L")
                ):
                    image = image.convert("RGB")
                image.save(buf, format, quality=90)
                raw_data = buf.getvalue()
                buf.close()

                storage.delete(miniature)
                storage.save(miniature, ContentFile(raw_data))

                image.close()


class CropscaleThumbnailer(Thumbnailer):
    THUMBNAIL_SIZE_RE = re.compile(r"^(?P<w>\d+)x(?P<h>\d+)(-(?P<x>\d+)x(?P<y>\d+))?$")
    MARKER = "_cropscale_"

    def generate(self, storage, original, size, miniature):
        with storage.open(original) as original_handle:
            with BytesIO(original_handle.read()) as original_bytes:
                image = Image.open(original_bytes)

                w, h = int(size["w"]), int(size["h"])

                if size["x"] and size["y"]:
                    x, y = int(size["x"]), int(size["y"])
                else:
                    x, y = 50, 50

                src_width, src_height = image.size
                src_ratio = float(src_width) / float(src_height)
                dst_width, dst_height = w, h
                dst_ratio = float(dst_width) / float(dst_height)

                if dst_ratio < src_ratio:
                    crop_height = src_height
                    crop_width = crop_height * dst_ratio
                    x_offset = int(float(src_width - crop_width) * x / 100)
                    y_offset = 0
                else:
                    crop_width = src_width
                    crop_height = crop_width / dst_ratio
                    x_offset = 0
                    y_offset = int(float(src_height - crop_height) * y / 100)

                assert image.format is not None
                format = image.format.lower()  # Save format for the save() call later
                image = image.crop(
                    (
                        x_offset,
                        y_offset,
                        x_offset + int(crop_width),
                        y_offset + int(crop_height),
                    )
                )
                image = image.resize((dst_width, dst_height), Image.Resampling.LANCZOS)

                buf = BytesIO()
                if format not in self.BROWSER_SUPPORTED_FORMATS:
                    format = "jpeg"
                if image.mode not in ("RGBA", "RGB", "L"):
                    if format in self.TRANSPARENCY_SUPPORTING_FORMATS:
                        image = image.convert("RGBA")
                    else:
                        image = image.convert("RGB")
                image.save(buf, format, quality=90)
                raw_data = buf.getvalue()
                buf.close()

                storage.delete(miniature)
                storage.save(miniature, ContentFile(raw_data))

                image.close()


@register.filter
def thumbnail(filename, size="200x200"):
    """
    Creates a thumbnail from the image passed, returning its path::

        {{ object.image|thumbnail:"400x300" }}
    OR
        {{ object.image.name|thumbnail:"400x300" }}

    You can pass either an ``ImageField``, ``FileField`` or the ``name``
    but not the ``url`` attribute of an ``ImageField`` or ``FileField``.

    The dimensions passed are treated as a bounding box. The aspect ratio of
    the initial image is preserved. Images aren't blown up in size if they
    are already smaller.

    Both width and height must be specified. If you do not care about one
    of them, just set it to an arbitrarily large number::

        {{ object.image|thumbnail:"300x999999" }}
    """

    return Thumbnailer(filename, size)


@register.filter
def cropscale(filename, size="200x200"):
    """
    Scales the image down and crops it so that its size equals exactly the size
    passed (as long as the initial image is bigger than the specification).
    """

    return CropscaleThumbnailer(filename, size)



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


import os
import re

from django.db import models
from django.db.models.signals import post_delete
from django.dispatch.dispatcher import receiver
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.models import ExtensionsMixin
from feincms.translations import (
    TranslatedObjectManager,
    TranslatedObjectMixin,
    Translation,
)

from . import logger


# ------------------------------------------------------------------------
class CategoryManager(models.Manager):
    """
    Simple manager which exists only to supply ``.select_related("parent")``
    on querysets since we can't even __str__ efficiently without it.
    """

    def get_queryset(self):
        return super().get_queryset().select_related("parent")


# ------------------------------------------------------------------------
class Category(models.Model):
    """
    These categories are meant primarily for organizing media files in the
    library.
    """

    title = models.CharField(_("title"), max_length=200)
    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="children",
        limit_choices_to={"parent__isnull": True},
        verbose_name=_("parent"),
    )

    slug = models.SlugField(_("slug"), max_length=150)

    class Meta:
        ordering = ["parent__title", "title"]
        verbose_name = _("category")
        verbose_name_plural = _("categories")
        app_label = "medialibrary"

    objects = CategoryManager()

    def __str__(self):
        if self.parent_id:
            return f"{self.parent.title} - {self.title}"

        return self.title

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.title)

        super().save(*args, **kwargs)

    save.alters_data = True

    def path_list(self):
        if self.parent is None:
            return [self]
        p = self.parent.path_list()
        p.append(self)
        return p

    def path(self):
        return " - ".join(f.title for f in self.path_list())


# ------------------------------------------------------------------------
class MediaFileBase(models.Model, ExtensionsMixin, TranslatedObjectMixin):
    """
    Abstract media file class. Includes the
    :class:`feincms.models.ExtensionsMixin` because of the (handy) extension
    mechanism.
    """

    file = models.FileField(
        _("file"), max_length=255, upload_to=settings.FEINCMS_MEDIALIBRARY_UPLOAD_TO
    )
    type = models.CharField(_("file type"), max_length=12, editable=False, choices=())
    created = models.DateTimeField(_("created"), editable=False, default=timezone.now)
    copyright = models.CharField(_("copyright"), max_length=200, blank=True)
    file_size = models.IntegerField(
        _("file size"), blank=True, null=True, editable=False
    )

    categories = models.ManyToManyField(
        Category, verbose_name=_("categories"), blank=True
    )
    categories.category_filter = True

    class Meta:
        abstract = True
        ordering = ["-created"]
        verbose_name = _("media file")
        verbose_name_plural = _("media files")

    objects = TranslatedObjectManager()

    filetypes = []
    filetypes_dict = {}

    @classmethod
    def reconfigure(cls, upload_to=None, storage=None):
        f = cls._meta.get_field("file")
        # Ugh. Copied relevant parts from django/db/models/fields/files.py
        # FileField.__init__ (around line 225)
        if storage:
            f.storage = storage
        if upload_to:
            f.upload_to = upload_to
            if callable(upload_to):
                f.generate_filename = upload_to

    @classmethod
    def register_filetypes(cls, *types):
        cls.filetypes[0:0] = types
        choices = [t[0:2] for t in cls.filetypes]
        cls.filetypes_dict = dict(choices)
        cls._meta.get_field("type").choices = choices

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.file:
            self._original_file_name = self.file.name

    def __str__(self):
        if settings.FEINCMS_MEDIAFILE_TRANSLATIONS:
            trans = None

            try:
                trans = self.translation
            except models.ObjectDoesNotExist:
                pass
            except AttributeError:
                pass

            if trans:
                trans = "%s" % trans
                if trans.strip():
                    return trans

        return os.path.basename(self.file.name)

    def get_absolute_url(self):
        return self.file.url

    def determine_file_type(self, name):
        """
        >>> str(MediaFile().determine_file_type('foobar.jpg'))
        'image'
        >>> str(MediaFile().determine_file_type('foobar.PDF'))
        'pdf'
        >>> str(MediaFile().determine_file_type('foobar.jpg.pdf'))
        'pdf'
        >>> str(MediaFile().determine_file_type('foobar.jgp'))
        'other'
        >>> str(MediaFile().determine_file_type('foobar-jpg'))
        'other'
        """
        for type_key, type_name, type_test in self.filetypes:
            if type_test(name):
                return type_key
        return self.filetypes[-1][0]

    def save(self, *args, **kwargs):
        if not self.id and not self.created:
            self.created = timezone.now()

        self.type = self.determine_file_type(self.file.name)
        if self.file:
            try:
                self.file_size = self.file.size
            except (OSError, ValueError) as e:
                logger.error(f"Unable to read file size for {self}: {e}")

        super().save(*args, **kwargs)

        logger.info(
            "Saved mediafile %d (%s, type %s, %d bytes)"
            % (self.id, self.file.name, self.type, self.file_size or 0)
        )

        # User uploaded a new file. Try to get rid of the old file in
        # storage, to avoid having orphaned files hanging around.
        if getattr(self, "_original_file_name", None):
            if self.file.name != self._original_file_name:
                self.delete_mediafile(self._original_file_name)

        self.purge_translation_cache()

    save.alters_data = True

    def delete_mediafile(self, name=None):
        if name is None:
            name = self.file.name
        try:
            self.file.storage.delete(name)
        except Exception as e:
            logger.warn(f"Cannot delete media file {name}: {e}")


# ------------------------------------------------------------------------
MediaFileBase.register_filetypes(
    # Should we be using imghdr.what instead of extension guessing?
    (
        "image",
        _("Image"),
        lambda f: re.compile(
            r"\.(bmp|jpe?g|jp2|jxr|gif|png|tiff?|webp)$", re.IGNORECASE
        ).search(f),
    ),
    (
        "video",
        _("Video"),
        lambda f: re.compile(
            r"\.(mov|m[14]v|mp4|avi|mpe?g|qt|ogv|wmv|flv)$", re.IGNORECASE
        ).search(f),
    ),
    (
        "audio",
        _("Audio"),
        lambda f: re.compile(r"\.(au|mp3|m4a|wma|oga|ram|wav)$", re.IGNORECASE).search(
            f
        ),
    ),
    ("pdf", _("PDF document"), lambda f: f.lower().endswith(".pdf")),
    ("swf", _("Flash"), lambda f: f.lower().endswith(".swf")),
    ("txt", _("Text"), lambda f: f.lower().endswith(".txt")),
    ("rtf", _("Rich Text"), lambda f: f.lower().endswith(".rtf")),
    ("zip", _("Zip archive"), lambda f: f.lower().endswith(".zip")),
    (
        "doc",
        _("Microsoft Word"),
        lambda f: re.compile(r"\.docx?$", re.IGNORECASE).search(f),
    ),
    (
        "xls",
        _("Microsoft Excel"),
        lambda f: re.compile(r"\.xlsx?$", re.IGNORECASE).search(f),
    ),
    (
        "ppt",
        _("Microsoft PowerPoint"),
        lambda f: re.compile(r"\.pptx?$", re.IGNORECASE).search(f),
    ),
    ("other", _("Binary"), lambda f: True),  # Must be last
)


# ------------------------------------------------------------------------
class MediaFile(MediaFileBase):
    class Meta:
        app_label = "medialibrary"


@receiver(post_delete, sender=MediaFile)
def _mediafile_post_delete(sender, instance, **kwargs):
    instance.delete_mediafile()
    logger.info("Deleted mediafile %d (%s)" % (instance.id, instance.file.name))


# ------------------------------------------------------------------------
class MediaFileTranslation(Translation(MediaFile)):
    """
    Translated media file caption and description.
    """

    caption = models.CharField(_("caption"), max_length=1000)
    description = models.TextField(_("description"), blank=True)

    class Meta:
        verbose_name = _("media file translation")
        verbose_name_plural = _("media file translations")
        unique_together = ("parent", "language_code")
        app_label = "medialibrary"

    def __str__(self):
        return self.caption



"""
Sometimes, a single title is not enough, you'd like subtitles, and maybe
differing titles in the navigation and in the <title>-tag.  This extension lets
you do that.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions
from feincms._internal import monkeypatch_property


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "_content_title",
            models.TextField(
                _("content title"),
                blank=True,
                help_text=_(
                    "The first line is the main title, the following"
                    " lines are subtitles."
                ),
            ),
        )

        self.model.add_to_class(
            "_page_title",
            models.CharField(
                _("page title"),
                max_length=69,
                blank=True,
                help_text=_(
                    "Page title for browser window. Same as title by"
                    " default. Must be 69 characters or fewer."
                ),
            ),
        )

        @monkeypatch_property(self.model)
        def page_title(self):
            """
            Use this for the browser window (<title>-tag in the <head> of the
            HTML document)
            """

            if self._page_title:
                return self._page_title
            return self.content_title

        @monkeypatch_property(self.model)
        def content_title(self):
            """
            This should be used f.e. for the <h1>-tag
            """

            if not self._content_title:
                return self.title

            return self._content_title.splitlines()[0]

        @monkeypatch_property(self.model)
        def content_subtitle(self):
            return "\n".join(self._content_title.splitlines()[1:])

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(
            _("Titles"),
            {"fields": ("_content_title", "_page_title"), "classes": ("collapse",)},
        )



from django.urls import path, re_path

from feincms.views import Handler


handler = Handler.as_view()

urlpatterns = [
    path("", handler, name="feincms_home"),
    re_path(r"^(.*)/$", handler, name="feincms_handler"),
]



import os

from django.conf import settings
from django.core.management.base import BaseCommand

from feincms.module.medialibrary.models import MediaFile


class Command(BaseCommand):
    help = "Prints all orphaned files in the `media/medialibrary` folder"

    def handle(self, **options):
        mediafiles = list(MediaFile.objects.values_list("file", flat=True))

        root_len = len(settings.MEDIA_ROOT)
        medialib_path = os.path.join(settings.MEDIA_ROOT, "medialibrary")

        for base, dirs, files in os.walk(medialib_path):
            for f in files:
                if base.startswith(settings.MEDIA_ROOT):
                    base = base[root_len:]
                full = os.path.join(base, f)
                if full not in mediafiles:
                    self.stdout.write(full)



"""
Add an excerpt field to the page.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "excerpt",
            models.TextField(
                _("excerpt"),
                blank=True,
                help_text=_(
                    "Add a brief excerpt summarizing the content" " of this page."
                ),
            ),
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(
            _("Excerpt"), {"fields": ("excerpt",), "classes": ("collapse",)}
        )



from .base import (
    Extension,
    ExtensionModelAdmin,
    ExtensionsMixin,
    prefetch_modeladmin_get_queryset,
)


__all__ = (
    "ExtensionsMixin",
    "Extension",
    "ExtensionModelAdmin",
    "prefetch_modeladmin_get_queryset",
)



from django.contrib.auth.models import User
from django.core.files import File as DjangoFile
from django.core.management.base import BaseCommand
from filer.models import File, Image

from feincms.contents import FilerFileContent, FilerImageContent
from feincms.module.medialibrary.contents import MediaFileContent
from feincms.module.medialibrary.models import MediaFile
from feincms.module.page.models import Page


PageMediaFileContent = Page.content_type_for(MediaFileContent)
PageFilerFileContent = Page.content_type_for(FilerFileContent)
PageFilerImageContent = Page.content_type_for(FilerImageContent)


assert all(
    (PageMediaFileContent, PageFilerFileContent, PageFilerImageContent)
), "Not all required models available"


class Command(BaseCommand):
    help = "Migrate the medialibrary and contents to django-filer"

    def handle(self, **options):
        user = User.objects.order_by("pk")[0]

        count = MediaFile.objects.count()

        for i, mediafile in enumerate(MediaFile.objects.order_by("pk")):
            model = Image if mediafile.type == "image" else File
            content_model = (
                PageFilerImageContent
                if mediafile.type == "image"
                else PageFilerFileContent
            )  # noqa

            filerfile = model.objects.create(
                owner=user,
                original_filename=mediafile.file.name,
                file=DjangoFile(mediafile.file.file, name=mediafile.file.name),
            )

            contents = PageMediaFileContent.objects.filter(mediafile=mediafile)

            for content in contents:
                content_model.objects.create(
                    parent=content.parent,
                    region=content.region,
                    ordering=content.ordering,
                    type=content.type,
                    mediafile=filerfile,
                )

                content.delete()

            if not i % 10:
                self.stdout.write(f"{i} / {count} files\n")

        self.stdout.write(f"{count} / {count} files\n")



VERSION = (25, 5, 1)
__version__ = ".".join(map(str, VERSION))


class LazySettings:
    def _load_settings(self):
        from django.conf import settings as django_settings

        from feincms import default_settings

        for key in dir(default_settings):
            if not key.startswith("FEINCMS_"):
                continue

            value = getattr(default_settings, key)
            value = getattr(django_settings, key, value)
            setattr(self, key, value)

    def __getattr__(self, attr):
        self._load_settings()
        del self.__class__.__getattr__
        return self.__dict__[attr]


settings = LazySettings()



from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms.admin.item_editor import FeinCMSInline
from feincms.utils.tuple import AutoRenderTuple


try:
    from filer.fields.file import FilerFileField
    from filer.fields.image import FilerImageField
except ImportError:
    __all__ = ()

else:
    __all__ = (
        "MediaFileContentInline",
        "ContentWithFilerFile",
        "FilerFileContent",
        "FilerImageContent",
    )

    class MediaFileContentInline(FeinCMSInline):
        radio_fields = {"type": admin.VERTICAL}

    class ContentWithFilerFile(models.Model):
        """
        File content
        """

        feincms_item_editor_inline = MediaFileContentInline

        class Meta:
            abstract = True

        def render(self, **kwargs):
            return AutoRenderTuple(
                (
                    [
                        f"content/filer/{self.file_type}_{self.type}.html",
                        "content/filer/%s.html" % self.type,
                        "content/filer/%s.html" % self.file_type,
                        "content/filer/default.html",
                    ],
                    {"content": self},
                )
            )

    class FilerFileContent(ContentWithFilerFile):
        mediafile = FilerFileField(
            verbose_name=_("file"), related_name="+", on_delete=models.CASCADE
        )
        file_type = "file"
        type = "download"

        class Meta:
            abstract = True
            verbose_name = _("file")
            verbose_name_plural = _("files")

    class FilerImageContent(ContentWithFilerFile):
        """
        Create a media file content as follows::

            from feincms.contents import FilerImageContent
            Page.create_content_type(FilerImageContent, TYPE_CHOICES=(
                ('inline', _('Default')),
                ('lightbox', _('Lightbox')),
                ('whatever', _('Whatever')),
            ))

        For a media file of type 'image' and type 'lightbox', the following
        templates are tried in order:

        * content/mediafile/image_lightbox.html
        * content/mediafile/lightbox.html
        * content/mediafile/image.html
        * content/mediafile/default.html

        The context contains ``content`` and ``request`` (if available).

        The content.mediafile attribute are as follows (selection):
        label, description, default_caption, default_alt_text,
        author, must_always_publish_author_credit,
        must_always_publish_copyright, date_taken, file, id, is_public, url
        """

        mediafile = FilerImageField(
            verbose_name=_("image"), related_name="+", on_delete=models.CASCADE
        )
        caption = models.CharField(_("caption"), max_length=1000, blank=True)
        url = models.CharField(_("URL"), max_length=1000, blank=True)

        file_type = "image"

        class Meta:
            abstract = True
            verbose_name = _("image")
            verbose_name_plural = _("images")

        @classmethod
        def initialize_type(cls, TYPE_CHOICES=None):
            if TYPE_CHOICES is None:
                raise ImproperlyConfigured(
                    "You have to set TYPE_CHOICES when" " creating a %s" % cls.__name__
                )

            cls.add_to_class(
                "type",
                models.CharField(
                    _("type"),
                    max_length=20,
                    choices=TYPE_CHOICES,
                    default=TYPE_CHOICES[0][0],
                ),
            )



"""
Simple image inclusion content: You should probably use the media library
instead.
"""


import os

from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.templatetags import feincms_thumbnail
from feincms.utils.tuple import AutoRenderTuple


class ImageContent(models.Model):
    # You should probably use
    # `feincms.content.medialibrary.models.MediaFileContent` instead.

    """
    Create an ImageContent like this::

        Cls.create_content_type(
            ImageContent,
            POSITION_CHOICES=(
                ('left', 'Float to left'),
                ('right', 'Float to right'),
                ('block', 'Block'),
            ),
            FORMAT_CHOICES=(
                ('noop', 'Do not resize'),
                ('cropscale:100x100', 'Square Thumbnail'),
                ('cropscale:200x450', 'Medium Portait'),
                ('thumbnail:1000x1000', 'Large'),
            ))

        Note that FORMAT_CHOICES is optional. The part before the colon
        corresponds to the template filters in the ``feincms_thumbnail``
        template filter library. Known values are ``cropscale`` and
        ``thumbnail``. Everything else (such as ``noop``) is ignored.
    """

    image = models.ImageField(
        _("image"),
        max_length=255,
        upload_to=os.path.join(settings.FEINCMS_UPLOAD_PREFIX, "imagecontent"),
    )
    alt_text = models.CharField(
        _("alternate text"),
        max_length=255,
        blank=True,
        help_text=_("Description of image"),
    )
    caption = models.CharField(_("caption"), max_length=255, blank=True)

    class Meta:
        abstract = True
        verbose_name = _("image")
        verbose_name_plural = _("images")

    def render(self, **kwargs):
        templates = ["content/image/default.html"]
        if hasattr(self, "position"):
            templates.insert(0, "content/image/%s.html" % self.position)

        return AutoRenderTuple((templates, {"content": self}))

    def get_image(self):
        img_type, _, size = getattr(self, "format", "").partition(":")
        if not size:
            return self.image

        thumbnailer = {"cropscale": feincms_thumbnail.CropscaleThumbnailer}.get(
            img_type, feincms_thumbnail.Thumbnailer
        )
        return thumbnailer(self.image, size)

    @classmethod
    def initialize_type(cls, POSITION_CHOICES=None, FORMAT_CHOICES=None):
        if POSITION_CHOICES:
            models.CharField(
                _("position"),
                max_length=10,
                choices=POSITION_CHOICES,
                default=POSITION_CHOICES[0][0],
            ).contribute_to_class(cls, "position")

        if FORMAT_CHOICES:
            models.CharField(
                _("format"),
                max_length=64,
                choices=FORMAT_CHOICES,
                default=FORMAT_CHOICES[0][0],
            ).contribute_to_class(cls, "format")



"""
Content types in FeinCMS 1.x returned a HTML fragment from ``render()``
methods.

FeinCMS 22.x changed them to return a ``(template_name, context)`` tuple. This
allows rendering content type templates while also having access to all
template variables set in views and context processors without having to
explicitly pass them on or something.

Some old projects still rely on ``render()`` returning a HTML string. This
tuple subclass combines the advantages of both approaches: Newer projects see a
tuple, old projects still get a string.
"""

from django.template.loader import render_to_string


class AutoRenderTuple(tuple):
    def __str__(self):
        return render_to_string(self[0], self[1])



from django.conf import settings as django_settings
from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.admin.item_editor import FeinCMSInline
from feincms.contrib.richtext import RichTextField
from feincms.module.medialibrary.fields import MediaFileForeignKey
from feincms.module.medialibrary.models import MediaFile
from feincms.utils.tuple import AutoRenderTuple


class SectionContentInline(FeinCMSInline):
    raw_id_fields = ("mediafile",)
    radio_fields = {"type": admin.VERTICAL}


class SectionContent(models.Model):
    """
    Title, media file and rich text fields in one content block.
    """

    feincms_item_editor_inline = SectionContentInline
    feincms_item_editor_context_processors = (
        lambda x: settings.FEINCMS_RICHTEXT_INIT_CONTEXT,
    )
    feincms_item_editor_includes = {"head": [settings.FEINCMS_RICHTEXT_INIT_TEMPLATE]}

    title = models.CharField(_("title"), max_length=200, blank=True)
    richtext = RichTextField(_("text"), blank=True)
    mediafile = MediaFileForeignKey(
        MediaFile,
        on_delete=models.CASCADE,
        verbose_name=_("media file"),
        related_name="+",
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True
        verbose_name = _("section")
        verbose_name_plural = _("sections")

    @classmethod
    def initialize_type(cls, TYPE_CHOICES=None, cleanse=None):
        if "feincms.module.medialibrary" not in django_settings.INSTALLED_APPS:
            raise ImproperlyConfigured(
                "You have to add 'feincms.module.medialibrary' to your"
                " INSTALLED_APPS before creating a %s" % cls.__name__
            )

        if TYPE_CHOICES is None:
            raise ImproperlyConfigured(
                "You need to set TYPE_CHOICES when creating a" " %s" % cls.__name__
            )

        cls.add_to_class(
            "type",
            models.CharField(
                _("type"),
                max_length=10,
                choices=TYPE_CHOICES,
                default=TYPE_CHOICES[0][0],
            ),
        )

        if cleanse:
            cls.cleanse = cleanse

    @classmethod
    def get_queryset(cls, filter_args):
        # Explicitly add nullable FK mediafile to minimize the DB query count
        return cls.objects.select_related("parent", "mediafile").filter(filter_args)

    def render(self, **kwargs):
        if self.mediafile:
            mediafile_type = self.mediafile.type
        else:
            mediafile_type = "nomedia"

        return AutoRenderTuple(
            (
                [
                    f"content/section/{mediafile_type}_{self.type}.html",
                    "content/section/%s.html" % mediafile_type,
                    "content/section/%s.html" % self.type,
                    "content/section/default.html",
                ],
                {"content": self},
            )
        )

    def save(self, *args, **kwargs):
        if getattr(self, "cleanse", None):
            try:
                # Passes the rich text content as first argument because
                # the passed callable has been converted into a bound method
                self.richtext = self.cleanse(self.richtext)
            except TypeError:
                # Call the original callable, does not pass the rich richtext
                # content instance along
                self.richtext = self.cleanse.im_func(self.richtext)

        super().save(*args, **kwargs)

    save.alters_data = True



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------


from functools import partial
from threading import local

from django.conf import settings as django_settings
from django.contrib import admin, messages
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.templatetags.static import static
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.admin import item_editor, tree_editor

# ------------------------------------------------------------------------
from .forms import PageAdminForm


_local = local()


# ------------------------------------------------------------------------
class PageAdmin(item_editor.ItemEditor, tree_editor.TreeEditor):
    class Media:
        css = {}
        js = []

    form = PageAdminForm

    fieldset_insertion_index = 2
    fieldsets = [
        (None, {"fields": [("title", "slug"), ("active", "in_navigation")]}),
        (
            _("Other options"),
            {
                "classes": ["collapse"],
                "fields": ["template_key", "parent", "override_url", "redirect_to"],
            },
        ),
        # <-- insertion point, extensions appear here, see insertion_index
        # above
        item_editor.FEINCMS_CONTENT_FIELDSET,
    ]
    readonly_fields = []
    list_display = [
        "short_title",
        "is_visible_admin",
        "in_navigation_toggle",
        "template",
    ]
    list_filter = ["active", "in_navigation", "template_key", "parent"]
    search_fields = ["title", "slug"]
    prepopulated_fields = {"slug": ("title",)}

    raw_id_fields = ["parent"]
    radio_fields = {"template_key": admin.HORIZONTAL}

    @classmethod
    def add_extension_options(cls, *f):
        if isinstance(f[-1], dict):  # called with a fieldset
            cls.fieldsets.insert(cls.fieldset_insertion_index, f)
            f[1]["classes"] = list(f[1].get("classes", []))
            f[1]["classes"].append("collapse")
        else:  # assume called with "other" fields
            cls.fieldsets[1][1]["fields"].extend(f)

    def __init__(self, model, admin_site):
        if len(model._feincms_templates) > 4 and "template_key" in self.radio_fields:
            del self.radio_fields["template_key"]

        super().__init__(model, admin_site)

    in_navigation_toggle = tree_editor.ajax_editable_boolean(
        "in_navigation", _("in navigation")
    )

    def get_readonly_fields(self, request, obj=None):
        readonly = super().get_readonly_fields(request, obj=obj)
        if not settings.FEINCMS_SINGLETON_TEMPLATE_CHANGE_ALLOWED:
            if obj and obj.template and obj.template.singleton:
                return tuple(readonly) + ("template_key",)
        return readonly

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        return partial(form, modeladmin=self)

    def _actions_column(self, page):
        addable = getattr(page, "feincms_addable", True)

        preview_url = "../../r/{}/{}/".format(
            ContentType.objects.get_for_model(self.model).id,
            page.id,
        )
        actions = super()._actions_column(page)

        if addable:
            if not page.template.enforce_leaf:
                actions.insert(
                    0,
                    '<a href="add/?parent=%s" title="%s">'
                    '<img src="%s" alt="%s" />'
                    "</a>"
                    % (
                        page.pk,
                        _("Add child page"),
                        static("feincms/img/icon_addlink.gif"),
                        _("Add child page"),
                    ),
                )
        actions.insert(
            0,
            '<a href="%s" title="%s">'
            '<img src="%s" alt="%s" />'
            "</a>"
            % (
                preview_url,
                _("View on site"),
                static("feincms/img/selector-search.gif"),
                _("View on site"),
            ),
        )
        return actions

    def add_view(self, request, **kwargs):
        kwargs["form_url"] = request.get_full_path()  # Preserve GET parameters
        if "translation_of" in request.GET and "language" in request.GET:
            try:
                original = self.model._tree_manager.get(
                    pk=request.GET.get("translation_of")
                )
            except (AttributeError, self.model.DoesNotExist):
                pass
            else:
                language_code = request.GET["language"]
                language = dict(django_settings.LANGUAGES).get(language_code, "")
                kwargs["extra_context"] = {
                    "adding_translation": True,
                    "title": _('Add %(language)s translation of "%(page)s"')
                    % {"language": language, "page": original},
                    "language_name": language,
                    "translation_of": original,
                }
        return super().add_view(request, **kwargs)

    def response_add(self, request, obj, *args, **kwargs):
        response = super().response_add(request, obj, *args, **kwargs)
        if (
            "parent" in request.GET
            and "_addanother" in request.POST
            and response.status_code in (301, 302)
        ):
            # Preserve GET parameters if we are about to add another page
            response["Location"] += "?parent=%s" % request.GET["parent"]

        if (
            "translation_of" in request.GET
            and "_copy_content_from_original" in request.POST
        ):
            # Copy all contents
            for content_type in obj._feincms_content_types:
                if content_type.objects.filter(parent=obj).exists():
                    # Short-circuit processing -- don't copy any contents if
                    # newly added object already has some
                    return response

            try:
                original = self.model._tree_manager.get(
                    pk=request.GET.get("translation_of")
                )
                original = original.original_translation
                obj.copy_content_from(original)
                obj.save()

                self.message_user(
                    request,
                    _(
                        "The content from the original translation has been copied"
                        " to the newly created page."
                    ),
                )
            except (AttributeError, self.model.DoesNotExist):
                pass

        return response

    def change_view(self, request, object_id, **kwargs):
        try:
            return super().change_view(request, object_id, **kwargs)
        except PermissionDenied:
            messages.add_message(
                request,
                messages.ERROR,
                _("You don't have the necessary permissions to edit this" " object"),
            )
        return HttpResponseRedirect(reverse("admin:page_page_changelist"))

    def has_delete_permission(self, request, obj=None):
        if not settings.FEINCMS_SINGLETON_TEMPLATE_DELETION_ALLOWED:
            if obj and obj.template.singleton:
                return False
        return super().has_delete_permission(request, obj=obj)

    def changelist_view(self, request, *args, **kwargs):
        _local.visible_pages = list(
            self.model.objects.active().values_list("id", flat=True)
        )
        return super().changelist_view(request, *args, **kwargs)

    @admin.display(description=_("is active"))
    def is_visible_admin(self, page):
        """
        Instead of just showing an on/off boolean, also indicate whether this
        page is not visible because of publishing dates or inherited status.
        """
        if page.parent_id and page.parent_id not in _local.visible_pages:
            # parent page's invisibility is inherited
            if page.id in _local.visible_pages:
                _local.visible_pages.remove(page.id)
            return tree_editor.ajax_editable_boolean_cell(
                page, "active", override=False, text=_("inherited")
            )

        if page.active and page.id not in _local.visible_pages:
            # is active but should not be shown, so visibility limited by
            # extension: show a "not active"
            return tree_editor.ajax_editable_boolean_cell(
                page, "active", override=False, text=_("extensions")
            )

        return tree_editor.ajax_editable_boolean_cell(page, "active")

    is_visible_admin.editable_boolean_field = "active"

    # active toggle needs more sophisticated result function
    def is_visible_recursive(self, page):
        # Have to refresh visible_pages here, because TreeEditor.toggle_boolean
        # will have changed the value when inside this code path.
        _local.visible_pages = list(
            self.model.objects.active().values_list("id", flat=True)
        )

        retval = []
        for c in page.get_descendants(include_self=True):
            retval.append(self.is_visible_admin(c))
        return retval

    is_visible_admin.editable_boolean_result = is_visible_recursive


# ------------------------------------------------------------------------
# ------------------------------------------------------------------------



# Thanks to http://www.djangosnippets.org/snippets/1051/
#
# Authors: Marinho Brandao <marinho at gmail.com>
#          Guilherme M. Gondim (semente) <semente at taurinus.org>


from operator import itemgetter

import django
from django.contrib.admin.filters import ChoicesFieldListFilter
from django.db.models import Count
from django.utils.encoding import smart_str
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from feincms.utils import shorten_string


class ParentFieldListFilter(ChoicesFieldListFilter):
    """
    Improved list_filter display for parent Pages by nicely indenting hierarchy

    In theory this would work with any mptt model which uses a "title"
    attribute.

    my_model_field.page_parent_filter = True
    """

    def __init__(self, field, request, params, model, model_admin, field_path=None):
        super().__init__(field, request, params, model, model_admin, field_path)

        parent_ids = (
            model.objects.exclude(parent=None)
            .values_list("parent__id", flat=True)
            .order_by("parent__id")
            .distinct()
        )
        parents = model.objects.filter(pk__in=parent_ids).values_list(
            "pk", "title", "level"
        )
        self.lookup_choices = [
            (
                pk,
                "{}{}".format(
                    "&nbsp;&nbsp;" * level, shorten_string(title, max_length=25)
                ),
            )
            for pk, title, level in parents
        ]

    def choices(self, changelist):
        yield {
            "selected": self.lookup_val is None,
            "query_string": changelist.get_query_string({}, [self.lookup_kwarg]),
            "display": _("All"),
        }

        # Pre Django 5 lookup_val would be a scalar, now it can do multiple
        # selections and thus is a list. Deal with that.
        lookup_vals = self.lookup_val
        if lookup_vals is not None and django.VERSION < (5,):
            lookup_vals = [lookup_vals]

        for pk, title in self.lookup_choices:
            yield {
                "selected": lookup_vals is not None and str(pk) in lookup_vals,
                "query_string": changelist.get_query_string({self.lookup_kwarg: pk}),
                "display": mark_safe(smart_str(title)),
            }

    def title(self):
        return _("Parent")


class CategoryFieldListFilter(ChoicesFieldListFilter):
    """
    Customization of ChoicesFilterSpec which sorts in the user-expected format

    my_model_field.category_filter = True
    """

    def __init__(self, field, *args, **kwargs):
        super().__init__(field, *args, **kwargs)

        # Restrict results to categories which are actually in use:
        related_model = field.remote_field.model
        related_name = field.related_query_name()

        self.lookup_choices = sorted(
            (
                (i.pk, f"{i} ({i._related_count})")
                for i in related_model.objects.annotate(
                    _related_count=Count(related_name)
                ).exclude(_related_count=0)
            ),
            key=itemgetter(1),
        )

    def choices(self, changelist):
        yield {
            "selected": self.lookup_val is None,
            "query_string": changelist.get_query_string({}, [self.lookup_kwarg]),
            "display": _("All"),
        }

        # Pre Django 5 lookup_val would be a scalar, now it can do multiple
        # selections and thus is a list. Deal with that.
        lookup_vals = self.lookup_val
        if lookup_vals is not None and django.VERSION < (5,):
            lookup_vals = [lookup_vals]

        for pk, title in self.lookup_choices:
            yield {
                "selected": lookup_vals is not None and str(pk) in lookup_vals,
                "query_string": changelist.get_query_string({self.lookup_kwarg: pk}),
                "display": mark_safe(smart_str(title)),
            }

    def title(self):
        return _("Category")



"""
Add a "featured" field to objects so admins can better direct top content.
"""


from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            "featured", models.BooleanField(_("featured"), default=False)
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(
            _("Featured"), {"fields": ("featured",), "classes": ("collapse",)}
        )



"""
Simple file inclusion content: You should probably use the media library
instead.
"""


import os

from django.db import models
from django.utils.translation import gettext_lazy as _

from feincms import settings
from feincms.utils.tuple import AutoRenderTuple


class FileContent(models.Model):
    # You should probably use
    # `feincms.content.medialibrary.models.MediaFileContent` instead.

    title = models.CharField(max_length=200)
    file = models.FileField(
        _("file"),
        max_length=255,
        upload_to=os.path.join(settings.FEINCMS_UPLOAD_PREFIX, "filecontent"),
    )

    class Meta:
        abstract = True
        verbose_name = _("file")
        verbose_name_plural = _("files")

    def render(self, **kwargs):
        return AutoRenderTuple(
            (
                ["content/file/%s.html" % self.region, "content/file/default.html"],
                {"content": self},
            )
        )



# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
"""
``rebuild_mptt``
---------------------

``rebuild_mptt`` rebuilds your mptt pointers. Only use in emergencies.
"""


from django.core.management.base import BaseCommand

from feincms.module.page.models import Page


class Command(BaseCommand):
    help = "Run this manually to rebuild your mptt pointers. Only use in emergencies."

    def handle_noargs(self, **options):
        self.handle(**options)

    def handle(self, **options):
        self.stdout.write("Rebuilding MPTT pointers for Page")
        Page._tree_manager.rebuild()
